#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2009 Efrim <efrim@azylum.org> {{{
# Copyright (c) 2009 J. Alexander Treuman (Tag collapse method)
# Copyright (c) 2008 Rick van Hattem (Track object)
#
#  This file is part of MPD_sima
#
#  MPD_sima is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation; either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public
#  License along with this program.
#  If not, see <http://www.gnu.org/licenses/>.
#
#
#  }}}

#{{{ DOC
"""
This code is dealing with your MPD server.
It will add automagicaly track to the playlist.
Simply run:
    python mpd_sima.py

See "python mpd_sima.py --help" for command line options.


This is a dev documentation. For user instructions please refer to doc/README.


 Unicode issue.
 --------------
I tried to use unicode everywhere I could but I still have all MPD returned
object UTF-8 formated. It leads to mixing unicode/UTF-8 strings all over the
code which is something I want to get rid of at some point.

    N.B. : MPD only deals with UTF-8


 The User DB and Cache
 ---------------------

USR_DB and CACHE are dict objects

USR_DB = {'artist O': [('similar 00', match00), ('similar 01', match01)],
          'artist 1': [('similar aa', matchaa), ('similar ab', matchab)]
          }
…
"""#}}}


__version__ = u'0.5.1'
__revison__ = u'$Revision: 216 $'[11:-2]
__author__ = u'$Author: kaliko $'
__date__ = u'$Date: 2009-11-26 14:51:38 +0100 (jeu 26 nov 2009) $'[7:26]

WAIT_MPD_RESUME = 9

#API_KEY = u'4a1c9ddec29816ed803d7be9113ba4cb'

# IMPORTS {{{
#import mpd
import random
import signal
import sys
import time

from collections import deque
from difflib import get_close_matches
#from os.path import (basename, isfile, getsize)
from socket import error as SocketError

from lib.audioscrobbler import (AudioScrobblerQuery,
        AudioScrobblerConnectionError)
from utils.config import (ConfMan)
from utils.leven import levenshtein_ratio
from utils.logutil import (logger, LEVELS)
from utils.startopt import StartOpt

try:
    from mpd import (MPDClient, ConnectionError, CommandError)
except ImportError, err:
    print 'ERROR: "%s"\n\nPlease install python-mpd module.\n' % err
    sys.exit(1)
#}}}

# OBJECTS{{{


class Track(object):#{{{
    """
    Track object.
    Instanciate with mpd replies.
    """

    def __init__(self, file=None, time=0, **kwargs):#{{{
        self.title = self.artist = self.album = ''
        self.pos = None
        self._file = file
        self.time = time
        self.__dict__.update(**kwargs)
        # Needed for multipule tags which returns a list instead of a string
        #  have tags been collapsed?
        self.collapse_tags_bool = False
        self.collapse_tags()#}}}

    def collapse_tags(self):#{{{
        """
        Necessary to deal with tags defined multiple times.
        """
        for tag, value in self.__dict__.iteritems():
            if isinstance(value, list):
                #self.__dict__[tag] = ", ".join(set(value))
                self.collapse_tags_bool = True
                self.__dict__.update({tag: ', '.join(set(value))})
                #print value#}}}

    def get_artist(self):#{{{
        """return lowercase UNICODE name of artist"""
        #return unicode(self.artist.lower(), 'utf-8')
        return unicode(self.artist, 'utf-8')

    def get_title(self):
        """return lowercase UNICODE title of song"""
        return unicode(self.title, 'utf-8')

    def get_album(self):
        """return lowercase UNICODE album of song"""
        return unicode(self.album, 'utf-8')

    def get_filename(self):
        """return filename"""
        return unicode(self.file, 'utf-8')

    def get_position(self):
        """return position of track in the playlist"""
        return int(self.pos)
    #}}}

    def __repr__(self):#{{{
        return '<%s: %s - %s - %s (%s)>' % (
            self.__class__.__name__,
            self.album,
            self.artist,
            self.title,
            self.duration,
        )

    def __int__(self):
        return self.time

    def __add__(self, other):
        return Track(time=self.time + other.time)

    def __sub__(self, other):
        return Track(time=self.time - other.time)

    def __hash__(self):
        if self.file:
            return hash(self.file)
        else:
            return id(self)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return hash(self) != hash(other)

    @property
    def file(self):
        """file is an immutable attribute that's used for the hash method"""
        return self._file

    def get_time(self):
        """get time property"""
        return self._time

    def set_time(self, value):
        """set time property"""
        self._time = int(value)

    time = property(get_time, set_time, doc='song duration in seconds')

    @property
    def duration(self):
        """Compute fancy duration"""
        temps = time.gmtime(int(self.time))
        if temps.tm_hour:
            fmt = '%H:%M:%S'
        else:
            fmt = '%M:%S'
        return time.strftime(fmt, temps)#}}}
#}}}


class Sima(object):#{{{
    """
    Main Object dealing with what and how to queue.
    """

    def __init__(self, config_man, log):#{{{
        """
        Declare default or empty attributes
        """
        self.current_track = None
        self.current_last_track = None
        self.current_queue = None
        #track object we are looking similar art/track for
        self.current_searched = None
        ## Cache
        self.cache_sim = config_man.similar_cache
        #self.cache_sim = dict()
        #self.cache_top = config_man.tops_cache
        self.cache_top = dict()
        self.cache_mpd = dict()
        ## Conf
        self.history = config_man.play_history
        self.config = config_man.config
        self.user_db = config_man.user_db
        self.conf_obj = config_man
        self.log = log
        ## MPD
        self.client = MPDClient()
        self.mpd_host = self.config.get('MPD', 'host')
        self.mpd_port = self.config.get('MPD', 'port')
        # TODO: Add method to get host/port/password
        # with password as boolean if not set?

        self.tracks_to_add = list()
        #}}}

    def mpd_connect(self):#{{{
        """
        Connection.
        """
        try:
            self.client.connect(self.mpd_host, self.mpd_port)
            return True
        except SocketError:
            return False#}}}

    def mpd_disconnect(self):#{{{
        """
        Wrapper around MPDClient().disconnect, intercepting errors.
        """
        try:
            self.client.disconnect()
            return True
        except (SocketError, ConnectionError):
            return False#}}}

    def mpd_auth(self):#{{{
        """
        Password auth.
        """
        try:
            self.config.getboolean('MPD', 'password')
            self.log.info(u'No password set, proceeding without ' +
                          u'authentication...')
        except ValueError:
            # ValueError if password not a boolean, hence an actual password.
            # pretty ugly? TODO: should I change this?
            try:
                self.client.password(self.config.get('MPD', 'password'))
                self.log.debug(u'Auth pass, proceeding...')
            except CommandError, err:
                self.log.error(u'Auth failed with "%s", wrong password?' % err)
                sys.exit(2)#}}}

    def mpd_current_song(self):#{{{
        """
        Currently played|paused|stopped track infos.
        """
        return Track(**self.client.currentsong())#}}}

    def mpd_findtrk(self, mpd_artists):#{{{
        """
        Find tracks to play (ie. not in history and etc.) while
        self.tracks_to_add is not reached.
        """
        self.tracks_to_add = []
        nbtracks_target = self.config.getint('sima', 'track_to_add')
        for artist in mpd_artists:
            mpd_find = [Track(**track) for track in self.client.find('artist', artist)]
            u_artist = unicode(artist, 'utf-8')
            self.log.debug(u'Trying to find titles to add for "%s"' %
                           u_artist)
            random.shuffle(mpd_find)
            unplayed_track = self._extract_unplayed(mpd_find)
            if not unplayed_track:
                self.log.info(u'Unable to find title to add' +
                              u' for "%s".' % u_artist)
            else:
                self.tracks_to_add.append(unplayed_track)
            if len(self.tracks_to_add) == nbtracks_target:
                break
        if not self.tracks_to_add:
            self.log.debug(u'Found no unplayed tracks, is your ' +
                             u'history getting large?')
            return False
        return True
        #}}}

    def mpd_find_top_tracks(self, mpd_artists):#{{{
        """
        Find top tracks for artists in mpd_artists list.
        """
        mpd_artists_u = [unicode(a, 'utf-8') for a in mpd_artists]
        self.tracks_to_add = list()
        nbtracks_target = self.config.getint('sima', 'track_to_add')
        ## DEBUG
        self.log.info(u'Looking for top tracks: "%s"...' %
                      u' / '.join(mpd_artists_u[0:4]))
        for artist in mpd_artists:
            if len(self.tracks_to_add) == nbtracks_target:
                return True
            titles_list = list()
            self.log.debug(u'Artist: "%s"' % unicode(artist, 'utf-8'))
            [titles_list.append(t) for t in self.get_top_tracks_from_db(artist)]

            for title in self.cross_check_titles(artist, titles_list):
                mpd_find = [Track(**t) for t in self.client.find('artist', artist, 'title', title)]
                unplayed_track = self._extract_unplayed(mpd_find)
                if not unplayed_track:
                    continue
                self.tracks_to_add.append(unplayed_track)
                break

        if not self.tracks_to_add:
            return False
        return True#}}}

    def mpd_add_track(self):#{{{
        """
        Add track to MPD.
        """
        ## Hight Verbosity DEBUG
        #self.log.debug(u'title history')
        #self.log.debug(u' - '.join([ a.get_artist() + ' ' +
        #               a.get_title() for a in self.history ]))
        ### END
        tracks = self.tracks_to_add
        for track in tracks:
            self.log.info(u'# Add to playlist: %s / %s' %
                          (track.get_artist(), track.get_title()))
            if not track.get_title():
                self.log.warning(u'MISSING TITLE TAG in: %s' %
                                    track.get_filename())
            ##DEV##ADD#
            try:
                self.client.add(track.file)
            except CommandError, err:
                self.log.warning(u'Cannot add track. (%s)' % err)
                msg = '[51@0] {add} playlist is at the max size'
                if str(err) in msg:
                    self.log.warning(u'MPD_sima hit playlist max size, '
                                     u'use consume mode or remove tracks.')
                return False
        #}}}

    def cross_check_titles(self, artist, titles):#{{{
        """
        cross_check_titles
        """
        u_artist = unicode(artist, 'utf-8')
        # Retrieve all tracks from artist
        all_tracks = self.client.find('artist', artist)
        # Get all utf8 titles (filter missing titles set to 'None')
        all_mpd_titles = [t.get('title') for t in all_tracks if t.get('title')]
        for title in titles:
            u_title = unicode(title, 'utf-8')
            # DEBUG
            #self.log.debug('looking for "%s" in MPD library.' % u_title)
            match = get_close_matches(title, all_mpd_titles, 50, 0.78)
            if not match:
                continue
            #self.log.debug(u'found close match for "%s": %s' %
            #               (u_title, match))
            for title_ in match:
                u_title_ = unicode(title_, 'utf-8')
                leven = levenshtein_ratio(title.lower(), title_.lower())
                if leven == 1:
                    yield title_
                    self.log.debug(u'"%s" matches "%s".' % (u_title_, u_title))
                elif leven >= 0.79:#PARAM
                    yield title_
                    self.log.debug(u'FZZZ: "%s" should match "%s" (lr=%1.3f)' %
                                   (u_title_, u_title, leven))
                else:
                    self.log.debug(u'FZZZ: "%s" does not match "%s" (lr=%1.3f)' %
                                   (u_title_, u_title, leven))
                    continue
        self.log.debug('Found no top tracks for "%s"' % u_artist)#}}}

    def cross_check_artist(self, artist_list):#{{{
        """
        # TODO: crosscheck even if exact match is returned !!!
        #       not straight forward, need heavy refactoring.
        # TODO: use unicode for artists lists is this method?
        """
        matching_artists = list()
        all_mpd_artists = self.client.list('artist')
        for artist in artist_list:
            u_artist = unicode(artist, 'utf-8')
            if artist in all_mpd_artists:
                matching_artists.append(artist)
                self.log.debug(u'found exact match for "%s"' % u_artist)
            else:
                match = get_close_matches(artist, all_mpd_artists, 50, 0.70)
                if match:
                    self.log.debug(u'found close match for "%s": %s' %
                                   (u_artist, match))
                for fuzz_art in match:
                    u_fuzz_art = unicode(fuzz_art, 'utf-8')
                    leven = levenshtein_ratio(artist.lower(), fuzz_art.lower())
                    if leven == 1:
                        matching_artists.append(fuzz_art)
                        self.log.debug(u'"%s" matches "%s".' %
                                       (u_fuzz_art, u_artist))
                    elif leven >= 0.79:#PARAM
                        matching_artists.append(fuzz_art)
                        self.log.debug(u'FZZZ: "%s" should match "%s" (lr=%1.3f)' %
                                       (u_fuzz_art, u_artist, leven))
                    else:
                        self.log.debug(u'FZZZ: "%s" does not match "%s" (lr=%1.3f)' %
                                       (u_fuzz_art, u_artist, leven))
            #else:
                #self.log.debug(u'Not found in MPD library: "%s"' % u_artist)
        return matching_artists#}}}

    def _extract_unplayed(self, tracks):#{{{
        """
        Extract one unplayed track from a Track object list.
        Check against history and file in queue.
        """
        track = None
        for track in tracks:
            if track not in self.tracks_to_add and not self._is_inqueue(track):
                if not self._is_inhist(track):
                    self.log.debug(u'track not yet in history: %s - %s' %
                                   (track.get_artist(), track.get_title()))
                    return track
                else:
                    self.log.debug(u'track already in history: %s - %s' %
                                   (track.get_artist(), track.get_title()))#}}}

    def _nb_track_ahead(self):#{{{
        """
        How many tracks ahead?
        """
        # current playing track position in the playlist & playlist length
        track_id = int(self.client.status().get('song', '0'))
        playlist_length = int(self.client.status().get('playlistlength', 0))
        return playlist_length - track_id -1#}}}

    def _get_last_track_inqueue(self):#{{{
        """
        TODO: find an alternate/elegant way to do so
        """
        last_track_pos = int(self.client.status().get('playlistlength', 0))-1
        self.current_last_track = Track(**self.client.playlistinfo(last_track_pos)[0])
        #}}}

    def _is_inqueue(self, track):#{{{
        """
        Check if track is in the queue.
        """
        cursonpos = int(self.client.currentsong().get('pos', 0))
        playlist = self.client.playlistinfo()
        # create a Track() list from the playlist queue
        queue_lst = [Track(**qtrack) for qtrack in playlist[cursonpos:]]
        if track in queue_lst:
            self.log.debug(u'"%s/%s/%s" already in the queue' %
                           (track.get_artist(), track.get_album(),
                            track.get_title()))
            return True
        return False#}}}

    def _is_inhist(self, track):#{{{
        """Check against history.
        """
        if track in self.history:
            return True
        return False#}}}

    def _need_tracks(self):#{{{
        """whether or not playlist needs tracks"""

        queue_trigger = self.config.getint('sima', 'queue_length')
        nb_track_ahead = self._nb_track_ahead()
        self.log.debug(u'Currently %i track(s) ahead. (target %s)' %
                       (nb_track_ahead, queue_trigger))
        if nb_track_ahead < queue_trigger:
            return True
        return False#}}}

    def _got_nothing(self):#{{{
        """log in case the script got nothing to add"""
        # TODO: use this method in get_artists_from_mpd() and queue()
        self.log.warning('Got nothing even with previous artists in playlist!')
        self.log.warning(u'...purge history?! rip more music?!')
        self.log.warning(u'Try running with debug verbosity to get more info.')
        #}}}

    def _get_artists_list_reorg(self, artists_list):#{{{
        """
        Move around items in artists_list in order to play not recently played
        artists

          N.B.  curr_play_hist is an UTF-8 list!!!!
                This is necessary to compare curr_play_hist w/ artists_list
                which is returned by MPD (hence UTF-8).

        TODO: add PARAM for curr_play_hist depth (ie. 20 currently)?
        TODO: while not PARAM for ↑ limit items fetch from self.history!!
        TODO: what have I done, list(deque([])) !??!?? (trying to understand
        with ipython, i came to the conclusion this is fucking useless…
        list(deque([])) == list([]). Need to run script to make shure I've
        actually wrote something stupid and correct it.
        """
        curr_play_hist = list(deque([a.artist for a in self.history]))
        nbtrkbkwd = len(artists_list) * 2
        #print "DEBUG: history - %s" % curr_play_hist[-nbtrkbkwd:]##DEBUG
        #print artists_list##DEBUG
        artists_list_pref = list(artists_list)
        for i in artists_list:
            #if i not in curr_play_hist[-nbtrkbkwd:]:
            if i not in curr_play_hist[-20:]:
                #print '%s not in recent history' % i##DEBUG
                artists_list_pref.remove(i)
                artists_list_pref.insert(0, i)
        #print artists_list_pref##DEBUG
        return artists_list_pref#}}}

    def get_top_tracks_from_db(self, artist=None):#{{{
        """
        Retrieve top tracks, ie. most popular song, from an artist.
        get_top_tracks_from_db function returns a list
        """
        tops = deque([])
        if artist in self.cache_top:
            tops = self.cache_top[artist]
            self.log.debug(u'Top tracks for "%s" already in cache, not requesting last.fm.' %
                           unicode(artist, 'utf-8'))
        else:
            try:
                req = AudioScrobblerQuery(artist=artist)
                tops = [str(song.name) for song in req.toptracks()]
                self.cache_top[artist] = tops
            except AudioScrobblerConnectionError, as_error:
                self.log.warning(u'Audioscrobbler server replies: %s...' %
                                    as_error)
        return tops#}}}

    def get_similar_artists_from_db(self):#{{{
        """
        Retrieve similar artists on last.fm server and from user DB if
        available.
        get_similar_artists_from_db function returns a list of tuple:
            [(artist, match),…].

        # N.B.
        #   <searched_artist> is UTF8.
        #   artist in the tuple is NOT UNICODE, this is probably UTF8…
        #   <usr_db> is also UTF8

        TODO: return unicode artists lists
        """
        similarity = self.config.getint('sima', 'similarity')
        searched_artist = self.current_searched.artist
        searched_artist_u = unicode(searched_artist, 'utf-8')
        self.log.info(u'Looking for artist similar to "%s"' %
                      searched_artist_u)

        # USER DB
        usr_db = self.user_db
        # DEBUG
        #usr_db = dict()
        # initialize artists deque list to construct from DB
        as_art = deque([])

        # Controls cache / append to cache
        #
        if searched_artist in self.cache_sim:
            as_art = self.cache_sim[searched_artist]
            self.log.debug(u'"%s" already in cache, not requesting last.fm.' %
                           searched_artist_u)
        else:
            as_artists = AudioScrobblerQuery(artist=searched_artist)
            self.log.debug(u'Requesting last.fm for "%s"' % searched_artist_u)
            try:
                [as_art.append((str(a.name), float(str(a.match)))) for a in as_artists.similar()]
            except AudioScrobblerConnectionError, err:
                self.log.warning(u'Server replies: ' % err)

            self.cache_sim[searched_artist] = as_art

        # Looks for user resources
        #
        # Append user_db entries to the list if
        #   searched_artist exist in user_db
        #    AND
        #   artists from user_db[searched_artist] not already in last.fm reply
        #   (n.b. reply == limited reply > similarity)

        # format last.fm artist names to compare last.fm replies with usr_db
        as_art_filt = [(aa.lower(), m) for aa, m in as_art if m > similarity]

        if searched_artist in usr_db:
            usr_db_art = usr_db[searched_artist]
            # Append user db entry if not already present in last.fm db
            # TODO: why not to use usr_db_art instead??!?
            appending = [a[0] for a in usr_db[searched_artist]]
            self.log.info(u'Checking “%s” from user DB.' %
                          unicode(' / '.join(appending), 'utf-8'))
            # extend artist list with artist from user db
            [as_art.extendleft((a,)) for a in usr_db_art if a not in as_art_filt]

        as_art_filt = [a for a, m in as_art if m > similarity]
        return as_art_filt#}}}

    def get_artists_from_mpd(self, artists_lst):#{{{
        """
        Look in MPD library for availability of similar artists retrieve from DB
        (DB: last.fm…)
        """
        similarity = self.config.getint('sima', 'similarity')
        artist = self.current_searched.artist
        mpd_artists = list()
        self.log.info(u'Looking availability in MPD library')
        if artist in self.cache_mpd:
            self.log.debug(u'Already cross check MPD library for this artist.')
            mpd_artists = self.cache_mpd.get(artist)
        else:
            mpd_artists = self.cross_check_artist(artists_lst)
            self.cache_mpd[artist] = mpd_artists
        if not mpd_artists:
            self._got_nothing()
            return None
        # Remove current artist in order to avoid loop. When the script is going
        # back in the playlist, because last searched does not return any track,
        # similar artist from DB does suggest the current artist which we
        # started similarity search with.

        ##DEBUG
        self.log.debug(u'current searched "%s"' %
                       self.current_track.get_artist())
        if self.current_track.artist in mpd_artists:
            self.log.info(u'Removing "%s" from artist list' %
                           self.current_track.get_artist())
            mpd_artists.remove(self.current_track.artist)
        self.log.info(u'Got %d artists in library (at least %d%% similar).' %
                      (len(mpd_artists), similarity))
        mpd_artists_u = [unicode(a, 'utf-8') for a in mpd_artists]
        self.log.info(u' / '.join(mpd_artists_u))
        # Look for tracks to add
        random.shuffle(mpd_artists)
        # Move around mpd_artists items to get in first
        # unplayed artist name. Sort of…
        return self._get_artists_list_reorg(mpd_artists)#}}}

    def queue_similar_artist(self):#{{{
        """
        Queue similar artist (at random)
        """
        similar = self.get_similar_artists_from_db()
        similar_u = [unicode(a, 'utf-8') for a in similar]
        if not similar:
            return False
        self.log.info(u'Some similar artist(s) – first five: %s...' %
                       u' / '.join(similar_u[0:5]))
        mpd_artists = self.get_artists_from_mpd(similar)
        if not mpd_artists:
            return False
        if not self.mpd_findtrk(mpd_artists):
            return False
        self.mpd_add_track()
        self.current_queue = self._nb_track_ahead()
        return True#}}}

    def queue_top_tracks(self):#{{{
        """
        Queue Top track from similar artist (at random)
        """
        similar = self.get_similar_artists_from_db()
        similar_u = [unicode(a, 'utf-8') for a in similar]
        if not similar:
            return False
        self.log.info(u'Some similar artist(s) – first five: %s...' %
                       u' / '.join(similar_u[0:5]))
        mpd_artists = self.get_artists_from_mpd(similar)
        if not mpd_artists:
            return False

        if not self.mpd_find_top_tracks(mpd_artists):
            return False
        self.mpd_add_track()
        self.current_queue = self._nb_track_ahead()
        return True#}}}

    def queue_mode(self):#{{{
        """
        Queue Mode:
            select right queue mode and return True/False
        """
        if self.config.getboolean('sima', 'top_tracks'):
            return self.queue_top_tracks()
        else:
            return self.queue_similar_artist()
        return False#}}}

    def queue(self, track):#{{{
        """
        On new track playing:
            add track in history
            Check either playlist needs more tracks or not.
            Find tracks to add.
        """
        if not track.artist:
            self.log.warning(u'## No artist tag set for %s' %
                             track.get_filename())
            self.log.warning(u'Cannont look for similar artist.')
            return False

        self.log.info(u'Playing: %s - %s' % (track.get_artist(),
                                           track.get_title()))
        if not self._need_tracks():
            return False

        self.log.info(u'The playlist needs tracks.')

        ##DEBUG
        #self.cache_mpd.clear()

        # Artist we want similar track from:
        #   currently played or last song in queue?
        # WARNING: changing from current to lastest leads to update the
        # current_* in get_artists_from_mpd() function where we look for its
        # presence in the mpd_artists list
        self.current_searched = self.current_track
        #self.current_searched = self.current_last_track

        # Already searched artists list (used when getting backward in play
        # history if nothing got queued)
        artist = self.current_searched.artist
        artists_searched = list([artist])

        history_copy = deque(self.history)
        while 42:
            if not self.queue_mode():
                # In case nothing got queued
                # get through play history backward until another artist got
                # something to queue
                self.log.debug('Looking for another artist in play history.')
                arthist = Track(artist=artist)
                while arthist.artist in artists_searched:
                    try:
                        arthist = history_copy.pop()
                        #DEBUG
                        #print arthist
                    except IndexError:
                        self._got_nothing()
                        return False
                # update the current_searched with new artist
                self.current_searched = arthist
                artists_searched.append(arthist.artist)
                #DEBUG
                #print artists_searched
                self.log.warning(u'Trying with previous artist: %s' %
                                self.current_searched.get_artist())
            else:
                break
        #}}}

    def connect(self):#{{{
        """
        MPD:
        Connection to mpd server.
        And controls available commands.
        """
        needed_cmds = ['status', 'add', 'find', \
                       'search', 'currentsong', 'ping']

        if not self.mpd_connect():
            self.log.error(u'Unable to connect to MPD on %s…' %
                    self.config.get('MPD', 'host'))
            return False
        self.log.info(u'Connected to MPD server (%s:%s), proceeding…' %
                         (self.mpd_host, self.mpd_port))
        self.mpd_auth()
        available_cmd = self.client.commands()
        for nddcmd in needed_cmds:
            if nddcmd not in available_cmd:
                self.log.error(u'command “%s” not available, '
                               u'control permissions… Need password?' %
                               nddcmd)
                sys.exit(2)
        return True#}}}

    def run(self):#{{{
        """
        Main Loop.
        Two events may trigger the queue process
            0) new track playing
            1) playing track has been moved or number of queued tracks has
               changed
        # TODO: use new idle command
        """
        self.connect()
        mpd_conn = True
        playing = True
        while 42:
            if mpd_conn:
                try:
                    curr_track = self.mpd_current_song()
                    # following used to detect deleted|moved tracks
                    curr_queue = self._nb_track_ahead()
                    mpd_state = str(self.client.status().get('state'))
                    if playing and mpd_state != 'play':
                        playing = False
                        self.log.info(u'MPD state is “%s” (check n*%is)' %
                                      (mpd_state, WAIT_MPD_RESUME))
                        time.sleep(WAIT_MPD_RESUME)
                        continue
                    elif not playing and mpd_state == 'play':
                        playing = True
                        self.log.info(u'Playing again, proceeding...')
                    if (curr_track != self.current_track or
                       curr_queue != self.current_queue) and playing:
                        if curr_track != self.current_track:
                            self.conf_obj.add_hist(curr_track)
                        # Update current playlist state
                        self.current_track = curr_track
                        self.current_queue = curr_queue
                        #self._get_last_track_inqueue()
                        ## DEBUG
                        #self.log.debug(curr_track)
                        self.queue(curr_track)
                except (ConnectionError, SocketError), err:
                    self.mpd_disconnect()
                    mpd_conn = False
                    self.log.warning(u'MPD connection lost.' +
                                     u'Trying to reconnect')
            elif not mpd_conn:
                if self.connect():
                    mpd_conn = True
                else:
                    self.log.warning(u'FAILED. Waiting %is to try again.' %
                                     WAIT_MPD_RESUME)
                    time.sleep(WAIT_MPD_RESUME)
            time.sleep(self.config.getint('sima', 'main_loop_time'))
        self.mpd_disconnect()
        #}}}
#}}}


# END OBJECTS }}}


# FUNCTIONS{{{


def sig_handler(signum, frame):#{{{
    """Catch signal other than KeyboardInterrupt"""
    raise KeyboardInterrupt(u'Caught a %d\' SIG TERM signal' % signum)#}}}


def shutdown(config_man):#{{{
    """Shutdown sequence."""
    config_man.cache_man()
    config_man.history_man()
    #}}}


def main(): #BOOT SEQUENCE {{{
    """
    Main function.
    """
    # Logging facility, default log level is INFO
    # DEBUG:
    #   force debug mode in the following loggers instance, remove to
    #   release !!!!!!
    #
    sopt = StartOpt(logger(log_level='info', name='boot'),
                    (__version__, __revison__, __date__))

    log = logger(log_level='debug', log_file=sopt.logfile)

    log.info(u'')
    log.info(u'Starting MPD_sima version %s (revision %s - %s)' %
             (__version__, __revison__, __date__))

    # Configuration manager Object
    conf_manager = ConfMan(log)
    config = conf_manager.config

    # Logging settings
    # Define the logger following user conf
    #  default log level is INFO.
    log.setLevel(LEVELS.get(config.get('log', 'verbosity')))

    # MPD connection
    log.debug(u'add %i tracks when playlist lenght <= %i track(s).' %
                      (config.getint('sima', 'track_to_add'),
                       config.getint('sima', 'queue_length')))
    log.debug(u'About to connect to %s:%d, password "%s"' %
            (config.get('MPD', 'host'), config.getint('MPD', 'port'),
            unicode(config.get('MPD', 'password'), 'utf-8')))


    sima = Sima(conf_manager, log)

    # In order to catch "kill 15" as KeyboardInterrupt when run on background
    signal.signal(signal.SIGTERM, sig_handler)

    try:
        sima.run()
    except KeyboardInterrupt, err:
        log.warning(u'Starting shutdown: %s' % err)
        shutdown(conf_manager)
        log.info(u'bye...')
        sys.exit(1)
    #}}}


# END FUNCTIONS}}}


# Script starts here
if __name__ == '__main__':
    main()

# VIM MODLINE
# vim: ai ts=4 sw=4 sts=4 expandtab
